from .field import Field, logger
import uuid

from ..redis_db import POPOTO_REDIS_DB


class KeyField(Field):
    """
    A KeyField is for unique identifiers and category searches
    All KeyFields are used for indexing on the DB.
    All keys together have unique_together enforced.

    UniqueKeyField and AutoKeyField provide unique constraint and auto-id generation respectively.
    All models must have one or more KeyFields.
    However an AutoKeyField will be automatically added to models without any specified KeyFields.

    todo: add support for https://github.com/ai/nanoid
    """
    unique: bool = False
    indexed: bool = False
    auto: bool = False
    auto_uuid_length: int = 32
    auto_id: str = ""
    null: bool = False
    max_length: int = 128

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        new_kwargs = {  # defaults
            'unique': True,
            'indexed': False,
            'auto': False,
            'auto_uuid_length': 32,
            'auto_id': "",
            'null': False,
            'max_length': 128,  # Redis limit is 512MB
        }
        new_kwargs.update(kwargs)
        for k in new_kwargs:
            setattr(self, k, new_kwargs[k])

        if self.auto:
            self.default = uuid.uuid4().hex[:self.auto_uuid_length]

    @classmethod
    def is_valid(cls, field, value) -> bool:
        if not super().is_valid(field, value):
            return False
        if field.auto and len(value) != field.auto_uuid_length:
            logger.error(f"auto key value is length {len(value)}. It should be {field.auto_uuid_length}")
            return False
        return True

    def get_filter_query_params(self, field_name: str) -> list:
        return super().get_filter_query_params(field_name) + [
            f'{field_name}',    # takes a str, exact match :x:
            f'{field_name}__contains',  # takes a str, matches :*x*:
            f'{field_name}__startswith',  # takes a str, matches :x*:
            f'{field_name}__endswith',  # takes a str, matches :*x:
            f'{field_name}__in',  # takes a list, returns any matches
        ]

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """

        keys_lists_to_intersect = list()
        db_key_length, field_key_position = model._meta.db_key_length, model._meta.get_db_key_index_position(field_name)
        num_keys_before = field_key_position
        num_keys_after = db_key_length - field_key_position - 1

        pipeline = POPOTO_REDIS_DB.pipeline()

        def get_key_pattern(query_value_pattern):
            key_pattern = f"{model._meta.db_class_key}:"
            key_pattern += ("*:" * num_keys_before)
            key_pattern += query_value_pattern
            key_pattern += ":*" * num_keys_after
            return key_pattern

        for query_param, query_value in query_params.items():

            if query_param.endswith('__in'):
                pipeline_2 = POPOTO_REDIS_DB.pipeline()
                for query_value_elem in query_value:
                    key_pattern = get_key_pattern(f"{query_value}")
                    pipeline_2 = pipeline_2.keys(key_pattern)
                keys_lists_to_union = pipeline_2.execute()
                keys_lists_to_intersect.append(set.union(*[set(key_list) for key_list in keys_lists_to_union]))

            else:
                if query_param == f'{field_name}':
                    key_pattern = get_key_pattern(f"{query_value}")
                elif query_param.endswith('__startswith'):
                    key_pattern = get_key_pattern(f"{query_value}*")
                elif query_param.endswith('__endswith'):
                    key_pattern = get_key_pattern(f"*{query_value}")

                pipeline = pipeline.keys(key_pattern)
            # todo: refactor to use HSCAN or sets, https://redis.io/commands/keys
            # https://redis-py.readthedocs.io/en/stable/index.html#redis.Redis.hscan_iter

        keys_lists_to_intersect += pipeline.execute()
        logger.debug(keys_lists_to_intersect)
        if len(keys_lists_to_intersect):
            return set.intersection(*[set(key_list) for key_list in keys_lists_to_intersect])
        return set()


class UniqueKeyField(KeyField):
    """
    UniqueKeyField() is equivalent to KeyField(unique=True)
    """
    def __init__(self, **kwargs):
        if not kwargs.get('unique', True):
            from ..models.base import ModelException
            raise ModelException("UniqueKey field MUST be unique")
        kwargs['unique'] = True
        super().__init__(**kwargs)


class AutoKeyField(UniqueKeyField):
    """
    AutoKeyField() is equivalent to KeyField(unique-True, auto=True)
    The AutoKeyField is an auto-generated, universally unique key
    It will be automatically added to models with no specified KeyFields
    Include this field in your model if you cannot otherwise enforce a unique-together constraint with other KeyFields.
    They auto-generated key is random and newly generated for a model instance.
    Model instances with otherwise identical properties are saved as separate instances with different auto-keys.
    """
    def __init__(self, **kwargs):
        kwargs['auto'] = True
        super().__init__(**kwargs)



import logging

logger = logging.getLogger('POPOTO.field')


class FieldBase(type):
    """Metaclass for all Popoto Fields."""

    def __new__(cls, name, bases, attrs, **kwargs):
        # if not a Field, skip setup
        parents = [b for b in bases if isinstance(b, FieldBase)]
        if not parents:
            return super().__new__(cls, name, bases, attrs, **kwargs)

        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        new_class.field_class_key = f"${name.strip('Field')}F"
        return new_class


class Field(metaclass=FieldBase):
    type: type = str
    unique: bool = False
    indexed: bool = False
    value: str = None
    null: bool = False
    max_length: int = 1024
    default: str = ""

    def __init__(self, **kwargs):
        field_options = {  # default
            'type': str,
            'unique': True,
            'indexed': False,
            'value': None,
            'null': False,
            'max_length': 1024,  # Redis limit is 512MB
            'default': None,
        }
        # set field_options, let kwargs override
        for k, v in field_options.items():
            setattr(self, k, kwargs.get(k, v))

    @classmethod
    def is_valid(cls, field, value) -> bool:
        if value is None and not field.null:
            logger.error(f"field {field} is null")
            return False
        if not field.null and not isinstance(value, field.type):
            logger.error(f"field {field} is type {field.type}. But value is {type(value)}")
            return False
        if field.type == str and len(str(value)) > field.max_length:
            return False
        return True

    @classmethod
    def format_value_pre_save(cls, field_value):
        """
        format field_value before saving to db
        return corrected field_value
        assumes validation is already passed
        """
        return field_value

    @classmethod
    def get_special_use_field_db_key(cls, model: 'Model', field_name: str):
        """
        For use by child class when implementing additional Redis data structures
        Children implementing more than one new structure will need to augment this.
        """
        return f"{cls.field_class_key}:{model._meta.db_class_key}:{field_name}"

    @classmethod
    def on_save(cls, model: 'Model', field_name: str, field_value, pipeline=None):
        from ..redis_db import POPOTO_REDIS_DB
        # todo: create indexes with Sets
        # if model._meta.fields[field_name].indexed:
        #     field_db_key = f"{cls.field_class_key}:{model._meta.db_class_key}:{field_name}"
        # #     this will not work! how to edit, delete, prevent overwrite and duplicates?
        #     field_value_b = cls.encode(field_value)
        #     if pipeline:
        #         return pipeline.set(field_db_key, field_value_b)
        #     else:
        #         return POPOTO_REDIS_DB.set(field_db_key, field_value_b)


    def get_filter_query_params(self, field_name: str) -> list:
        # todo: auto manage sets of db_keys to allow filter by any indexed field
        if self.indexed:
            params = [f'{field_name}', ]
            if self.null:
                params += [f'{field_name}__isnull', ]
            return params
        return list()

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """
        from ..models.query import QueryException
        if model._meta.fields[field_name].indexed:
            raise QueryException("Query filter by any indexed field is not yet implemented.")
        else:
            raise QueryException("Query filter not possible on non-indexed fields")



from datetime import datetime

import msgpack
import redis

import logging

from .query import Query
from ..fields.key_field import KeyField
from ..fields.field import Field
from ..fields.sorted_field import SortedField
from ..redis_db import POPOTO_REDIS_DB, ENCODING

logger = logging.getLogger('POPOTO.model_base')


class ModelException(Exception):
    pass


class ModelOptions:
    def __init__(self, model_name):
        self.model_name = model_name
        self.hidden_fields = {}
        self.explicit_fields = {}
        self.key_field_names = []
        self.list_field_names = []
        # self.set_field_names = []
        self.sorted_field_names = []
        self.geo_field_names = []
        self.indexed_field_names = []

        # todo: should this be a dict of related objects or just a list of field names?
        # self.related_fields = {}  # model becomes graph node

        # todo: allow customizing this in model.Meta class
        self.db_class_key = self.model_name

        self.abstract = False
        self.unique_together = []
        self.index_together = []
        self.parents = []
        self.auto_created = False
        self.base_meta = None

    def add_field(self, field_name: str, field: Field):
        if field_name.startswith("_") and field_name not in self.hidden_fields:
            self.hidden_fields[field_name] = field
        elif field_name not in self.explicit_fields:
            self.explicit_fields[field_name] = field
        else:
            raise ModelException(f"{field_name} is already a Field on the model")

        if field.indexed:
            self.indexed_field_names.append(field_name)

        if isinstance(field, KeyField):
            self.key_field_names.append(field_name)
        elif isinstance(field, SortedField):
            self.sorted_field_names.append(field_name)
        # elif isinstance(field, ListField):
        #     self.list_field_names.append(field_name)
        # elif isinstance(field, GeoField):
        #     self.geo_field_names.append(field_name)

    @property
    def fields(self) -> dict:
        return {**self.explicit_fields, **self.hidden_fields}

    @property
    def field_names(self) -> list:
        return list(self.fields.keys())

    @property
    def db_key_length(self):
        return 1 + len(self.key_field_names)

    def get_db_key_index_position(self, field_name):
        return 1 + sorted(self.key_field_names).index(field_name)


class ModelBase(type):
    """Metaclass for all Popoto Models."""

    def __new__(cls, name, bases, attrs, **kwargs):

        # Initialization is only performed for a Model and its subclasses
        parents = [b for b in bases if isinstance(b, ModelBase)]
        if not parents:
            return super().__new__(cls, name, bases, attrs, **kwargs)

        # logger.debug({k: v for k, v in attrs.items() if not k.startswith('__')})
        module = attrs.pop('__module__')
        new_attrs = {'__module__': module}
        attr_meta = attrs.pop('Meta', None)
        options = ModelOptions(name)
        options.parents = parents

        for obj_name, obj in attrs.items():
            if obj_name.startswith("__"):
                # builtin or inherited private vars and methods
                new_attrs[obj_name] = obj

            elif isinstance(obj, Field):
                # save field instance
                # attr will be overwritten as a field.type
                # model will handle this and set default values
                options.add_field(obj_name, obj)

            elif callable(obj) or hasattr(obj, '__func__') or hasattr(obj, '__set__'):
                # a callable method or property
                new_attrs[obj_name] = obj

            elif obj_name.startswith("_"):
                # a private static attr not to be saved in the db
                new_attrs[obj_name] = obj

            else:
                raise ModelException(
                    f"public model attributes must inherit from class Field. "
                    f"Try using a private var (eg. _{obj_name})_"
                )

        # todo: handle multiple inheritance
        # for base in parents:
        #     for field_name, field in base.auto_fields.items():
        #         options.add_field(field_name, field)

        new_class = super().__new__(cls, name, bases, new_attrs)

        options.abstract = getattr(attr_meta, 'abstract', False)
        options.meta = attr_meta or getattr(new_class, 'Meta', None)
        options.base_meta = getattr(new_class, '_meta', None)
        new_class._meta = options
        setattr(new_class, 'query', Query(new_class))
        setattr(new_class, 'objects', new_class.query)
        return new_class


class Model(metaclass=ModelBase):
    query: Query

    def __init__(self, **kwargs):
        cls = self.__class__
        # self._ttl = kwargs.get('ttl', None)
        # self._expire_at = kwargs.get('expire_at', None)

        # allow init kwargs to set any base parameters
        self.__dict__.update(kwargs)

        # add auto KeyField if needed
        if not len(self._meta.key_field_names):
            self._meta.add_field('_auto_key', KeyField(auto=True, key=cls.__name__))

        # set defaults
        for field_name, field in self._meta.fields.items():
            setattr(self, field_name, field.default)

        # set field values from init kwargs
        for field_name in self._meta.fields.keys() & kwargs.keys():
            setattr(self, field_name, kwargs.get(field_name))

        # todo: handle some key management?
        # perhaps need to prepare to handle events such as key change, partial key search, ...

        self._ttl = None  # todo: set default in child Meta class
        self._expire_at = None  # todo: datetime? or timestamp?

        # validate initial attributes
        if not self.is_valid(null_check=False):  # exclude null, will validate null values on pre-save
            raise ModelException(f"Could not instantiate class {self}")
        self._db_content = dict()  # empty until self.load_from_db() or self.save() called

    @property
    def db_key(self):
        """
        the db key must include the class name - equivalent to db table name
        like it or not, keys append alphabetically.
        if needed, can propose feature to include order=int param on KeyField to force an order
        """
        return f"{self._meta.db_class_key}:" + ":".join([
            getattr(self, key_field_name) for key_field_name in sorted(self._meta.key_field_names)
        ])

    def __repr__(self):
        return str({k: v for k, v in self.__dict__.items() if k in self._meta.fields})

    def __str__(self):
        return str(self.db_key)

    def __eq__(self, other):
        """
        equality method
        instances with the same key(s) and class are considered equal
        except when any key(s) are None, they are not equal to anything except themselves.

        for evaluating all instance values against each other, use something like this:
        self_dict = self._meta.fields.update((k, self.__dict__[k]) for k in set(self.__dict__).intersection(self._meta.fields))
        other_dict = other._meta.fields.update((k, other.__dict__[k]) for k in set(other.__dict__).intersection(other._meta.fields))
        return repr(dict(sorted(self_dict))) == repr(dict(sorted(other_dict)))
        """
        try:
            if isinstance(other, self.__class__):
                # always False if if any KeyFields are None
                if (None in [self._meta.fields.get(key_field_name) for key_field_name in self._meta.key_field_names]) or \
                        (None in [other._meta.fields.get(key_field_name) for key_field_name in other._meta.key_field_names]):
                    return repr(self) == repr(other)
                return self.db_key == other.db_key
        except:
            return False
        else:
            return False

    # @property
    # def field_names(self):
    #     return [
    #         k for k, v in self.__dict__.items()
    #         if all([not k.startswith("_"), k + "_meta" in self.__dict__])
    #     ]

    def is_valid(self, null_check=True) -> bool:
        """
            todo: validate values
            - field.type ✅
            - field.null ✅
            - field.max_length ✅
            - ttl, expire_at - todo
        """

        for field_name in self._meta.field_names:
            # type check the field values against their class specified type, unless null/None

            if all([
                getattr(self, field_name) is not None,
                not isinstance(getattr(self, field_name), self._meta.fields[field_name].type)
            ]):
                try:
                    setattr(self, field_name, self._meta.fields[field_name].type(getattr(self, field_name)))
                    if not isinstance(getattr(self, field_name), self._meta.fields[field_name].type):
                        raise TypeError(f"{field_name} is not type {self._meta.fields[field_name].type}. ")
                except TypeError as e:
                    logger.error(
                        f"{str(e)} \n Change the value or modify type on {self.__class__.__name__}.{field_name}"
                    )
                    return False

            # check non-nullable fields
            if null_check and \
                    self._meta.fields[field_name].null is False and \
                    getattr(self, field_name) is None:
                error = f"{field_name} is None/null. Set a value or set null=True on {self.__class__.__name__}.{field_name}"
                logger.error(error)
                return False

            # validate str max_length
            if self._meta.fields[field_name].type == str and \
                    getattr(self, field_name) and \
                    len(getattr(self, field_name)) > self._meta.fields[field_name].max_length:
                error = f"{field_name} is greater than max_length={self._meta.fields[field_name].max_length}"


            if self._ttl and self._expire_at:
                raise ModelException("Can set either ttl and expire_at. Not both.")

        for field_name, field_value in self.__dict__.items():
            if field_name in self._meta.fields.keys():
                field_class = self._meta.fields[field_name].__class__
                if not field_class.is_valid(self._meta.fields[field_name], field_value):
                    error = f"Validation on [{field_name}] Field failed"
                    logger.error(error)
                    return False

        return True

    def save(self, pipeline: redis.client.Pipeline = None, ttl=None, expire_at=None, ignore_errors: bool = False, *args,
             **kwargs):
        """
            Default save method. Uses Redis HSET command with key, dict of values, ttl.
        """
        if not self.is_valid():
            error_message = "Model instance parameters invalid. Failed to save."
            if ignore_errors:
                logger.error(error_message)
            else:
                raise ModelException(error_message)
            return pipeline or 0

        ttl, expire_at = (ttl or self._ttl), (expire_at or self._expire_at)

        # run any necessary formatting on field data before saving
        for field_name, field in self._meta.fields.items():
            setattr(
                self, field_name,
                field.format_value_pre_save(getattr(self, field_name))
            )

        hset_mapping = self.encode()
        self._db_content = hset_mapping

        for field_name, field in self._meta.fields.items():
            if pipeline:
                pipeline = field.on_save(
                    self, field_name=field_name,
                    field_value=getattr(self, field_name),
                    pipeline=pipeline
                )
            else:
                db_response = field.on_save(
                    self, field_name=field_name,
                    field_value=getattr(self, field_name)
                )

        if isinstance(pipeline, redis.client.Pipeline):
            pipeline = pipeline.hset(self.db_key, mapping=hset_mapping)
            pipeline = pipeline if ttl is None else pipeline.expire(self.db_key, ttl)
            pipeline = pipeline if expire_at is None else pipeline.expire_at(self.db_key, expire_at)
            return pipeline
        else:
            db_response = POPOTO_REDIS_DB.hset(self.db_key, mapping=hset_mapping)
            if ttl is not None: POPOTO_REDIS_DB.expire(self.db_key, ttl)
            if expire_at is not None: POPOTO_REDIS_DB.expireat(self.db_key, ttl)
            return db_response

    @classmethod
    def create(cls, *args, **kwargs):
        instance = cls(*args, **kwargs)
        instance.save()
        return instance

    @classmethod
    def get(cls, db_key: str = None, **kwargs):
        return cls.query.get(db_key=db_key, **kwargs)

    def encode(self):
        from .encoding import encode_popoto_model_obj
        return encode_popoto_model_obj(self)

    def load_from_db(self):
        self._db_content = POPOTO_REDIS_DB.hgetall(self.db_key)
        for key_b, db_value_b in self._db_content.items():
            setattr(self, key_b.decode("utf-8"), msgpack.unpackb(db_value_b))

    def revert(self):
        self.load_from_db()

    def delete(self, pipeline=None, *args, **kwargs):
        self._db_content = dict()
        if pipeline is not None:
            pipeline = pipeline.delete(self.db_key)
            return pipeline
        else:
            db_response = POPOTO_REDIS_DB.delete(self.db_key)
            if db_response >= 0:
                return True



import logging
import numpy as np
import pandas as pd
import redis.client

from src.popoto._archive.key_value import KeyValueModel
from ..redis_db import POPOTO_REDIS_DB
from src.popoto.exceptions import ModelException
logger = logging.getLogger(__name__)


class SortedSetException(ModelException):
    pass


class SortedSetModel(KeyValueModel):
    """
    stores things in a sorted set
    todo: split the db by each publisher source
    """
    class_describer = "sortedset"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # 'sort_score' REQUIRED, VALIDATE
        try:
            self.sort_score = int(kwargs['score'])  # int eg. 5
        except KeyError:
            raise SortedSetException("score required for SortedSetModel objects")
        except ValueError:
            raise SortedSetException(f"score must be castable as integer, received {kwargs.get('score')}")
        except Exception as e:
            raise SortedSetException(str(e))

    def save_own_existance(self, describer_key=""):
        self.describer_key = describer_key or f'{self.__class__.class_describer}:{self._db_key}'


    @classmethod
    def query(cls, key: str = "", key_suffix: str = "", key_prefix: str = "",
              score: int = None,
              score_range: float = 1,
              *args, **kwargs) -> dict:
        """
        :param key: the exact redis sortedset key (optional)
        :param key_suffix: suffix on the key  (optional)
        :param key_prefix: prefix on the key (optional)
        :param score: score for most recent value returned  (optional, default returns latest)
        :param score_range: number of periods desired in results (optional, default 0, so only return 1 value)
        :return: dict(values=[], ...)
        """

        sorted_set_key = cls.compile_db_key(key=key, key_prefix=key_prefix, key_suffix=key_suffix)
        # logger.debug(f'query for sorted set key {sorted_set_key}')
        # example key f'{key_prefix}:{cls.__name__}:{key_suffix}'

        # do a quick check to make sure this is a class of things we know is in existence
        describer_key = f'{cls.class_describer}:{sorted_set_key}'
        # if no score, assume query to find the most recent, the last one

        if not score:
            query_response = POPOTO_REDIS_DB.zrange(sorted_set_key, -1, -1)
            try:
                [value, score] = query_response[0].decode("utf-8").split(":")
            except:
                value, score = "unknown", 0

            min_score = max_score = score

        else:
            target_score = score
            logger.debug(f"querying for key {sorted_set_key} with score {target_score} and {score_range} backward")

            min_score, max_score = (target_score - score_range), target_score

            query_response = POPOTO_REDIS_DB.zrangebyscore(sorted_set_key, min_score, max_score)

        # NEW example query_response = [b'100:1']
        # which came from f'{self.value}:{str(score)}' where score = self.score

        return_dict = {
            'values': [],
            'values_count': 0,
            'score': score,
            'earliest_score': min_score,
            'latest_score': max_score,
            'score_range': score_range,
        }

        if not len(query_response):
            return return_dict

        try:
            return_dict['values_count'] = len(query_response)

            if len(query_response) < score_range + 1:
                return_dict["warning"] = "fewer values than query's periods_range"

            values = [value_score.decode("utf-8").split(":")[0] for value_score in query_response]
            scores = [float(value_score.decode("utf-8").split(":")[1]) for value_score in query_response]
            # todo: double check that [-1] in list is most recent score

            return_dict.update({
                'values': values,
                'scores': scores,
                'earliest_score': scores[0],
                'latest_score': scores[-1],
            })
            return return_dict

        except IndexError:
            return return_dict

        except Exception as e:
            logger.error("redis query problem: " + str(e))
            return {'error': "redis query problem: " + str(e),  # wtf happened?
                    'values': []}

    @staticmethod
    def get_values_array_from_query(query_results: dict, limit: int = 0):

        value_array = [float(v) for v in query_results['values']]

        if limit:
            if not isinstance(limit, int) or limit < 1:
                raise SortedSetException(f"bad limit: {limit}")

            elif len(value_array) > limit:
                value_array = value_array[-limit:]

        return np.array(value_array)

    def get_z_add_data(self, dataframe: pd.DataFrame = None):
        return {
            "key": self._db_key,
            "name": dataframe.to_dict() if dataframe else f'{self.value}:{self.score}',
            "score": self.score
        }

    def save(self, publish: bool = False, pipeline: redis.client.Pipeline = None, dataframe: pd.DataFrame = None, *args, **kwargs):
        if self.value is None and dataframe is None:
            raise ModelException("no value set, nothing to save!")

        self.save_own_existance()

        z_add_data = self.get_z_add_data(dataframe=dataframe)

        if isinstance(pipeline, redis.client.Pipeline):
            pipeline = pipeline.zadd(z_add_data["key"], z_add_data["name"])
            # logger.debug("added command to redis pipeline")
            if publish:
                pipeline = self.publish(pipeline)
            return pipeline
        else:
            response = POPOTO_REDIS_DB.zadd(z_add_data["key"], z_add_data["name"])
            # logger.debug("no pipeline, executing zadd command immediately.")
            if publish:
                self.publish()
            return response

    def publish(self, pipeline=None):
        return super().publish(data=self.get_z_add_data(), pipeline=pipeline)

    def get_value(self, *args, **kwargs):
        SortedSetException("function not yet implemented! ¯\_(ツ)_/¯ ")
        pass


"""
We can scan the newest or oldest event ids with ZRANGE 4,
maybe later pulling the events themselves for analysis.

We can get the 10 or even 100 events immediately
before or after a score with ZRANGEBYSCORE
combined with the LIMIT argument.

We can count the number of events that occurred
in a specific score range with ZCOUNT.
"""



import msgpack
import datetime
from ..fields.geo_field import GeoField

from ..redis_db import ENCODING


def encode_custom_types(obj):
    if isinstance(obj, GeoField.Coordinates):
        return {'__Coordinates__': True, 'latitude': obj.latitude, 'longitude': obj.longitude}
    if isinstance(obj, datetime.datetime):
        return {'__datetime__': True, 'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f")}
    if isinstance(obj, datetime.date):
        return {'__date__': True, 'as_str': obj.strftime("%Y%m%d")}
    if isinstance(obj, datetime.time):
        return {'__time__': True, 'as_str': obj.strftime("%H:%M:%S.%f")}
    return obj


def decode_custom_types(obj):
    if isinstance(obj, dict) and 'as_str' in obj:
        if '__Coordinates__' in obj:
            return GeoField.Coordinates(obj.latitude, obj.longitude)
        if '__datetime__' in obj:
            return datetime.datetime.strptime(obj['as_str'], "%Y%m%dT%H:%M:%S.%f")
        if '__date__' in obj:
            return datetime.datetime.strptime(obj['as_str'], "%Y%m%d").date()
        if '__time__' in obj:
            return datetime.datetime.strptime(obj['as_str'], "%H:%M:%S.%f").time()
    return obj


def encode_popoto_model_obj(obj: 'Model') -> dict:
    import msgpack_numpy as m
    m.patch()

    encoded_hashmap = dict()
    for field_name, field in obj._meta.fields.items():
        value = getattr(obj, field_name)

        if field.type in [GeoField.Coordinates, datetime.datetime, datetime.date, datetime.time]:
            encoded_value = msgpack.packb(value, default=encode_custom_types)
        else:
            encoded_value = msgpack.packb(value)

        encoded_hashmap[str(field_name).encode(ENCODING)] = encoded_value

    return encoded_hashmap


def decode_popoto_model_hashmap(model_class: 'Model', redis_hash: dict) -> 'Model':
    return model_class(**{
        key_b.decode("utf-8"): decode_custom_types(msgpack.unpackb(value_b))
        for key_b, value_b in redis_hash.items()
    })



from .fields.field import Field
from .fields.key_field import KeyField, UniqueKeyField, AutoKeyField
from .fields.geo_field import GeoField
from .fields.sorted_field import SortedField
from .models.base import Model, ModelBase
from .pubsub.publisher import Publisher
from .pubsub.subscriber import Subscriber

__all__ = [
    'Field',
    'KeyField', 'UniqueKeyField', 'AutoKeyField',
    'GeoField',
    'SortedField',
    'Model', 'ModelBase',
    'Publisher', 'Subscriber',
]
